#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <mpi.h>
#include "functions.h"

#ifndef DEBUG
#define PrintPic(picptr) {}
#endif

#define Master 0
#define buffer_t char

#define TASK_NUM 4

MPI_Status status;
int myRank, nTasks;
buffer_t* buffer;
MPI_Datatype MPI_Answer;

int main(int Argc, char *Argv[])
{
	typedef struct {
		int Compare;
		int CountChar;
	}
	Answer_t;
	Answer_t *Answer;

	char **pic = dummy;
	char **picptr = NULL;
	char *ObjectNames[] = { "Kreis   ", "Dreieck ", "Rechteck"};

	int i = 0, j;
	int me;
	int k, l, m;
	double DTime, Time;
	int columns, rows, size;
	int MinCompare, MinCountChar;
	int MinVerIndex, MinCountIndex;

	MPI_Init(&Argc, &Argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &myRank);
	MPI_Comm_size(MPI_COMM_WORLD, &nTasks);

	srand(44444);
	if (nTasks < TASK_NUM) {
		if (myRank == Master) {
			printf("Programm sollte auf %d Knoten laufen\n", TASK_NUM);
		}
		MPI_Finalize();
		
		return 1;
	} else {
		rows = GetRows(pic);
		columns = GetColumns(pic);
		size = (rows * columns) + 1;
		
		buffer = (buffer_t*)malloc(size * sizeof(buffer_t));
		
		if (myRank == Master) {
			Answer = (Answer_t*)malloc(sizeof(Answer_t) * nTasks);
		} else {
			Answer = (Answer_t*)malloc(sizeof(Answer_t) * 1);
			me = (myRank - 1) % (TASK_NUM - 1);
			
			pic = (char**)malloc(sizeof(void*) * (columns + 1));
			for (l = 0; l < columns; l++) {
				pic[l] = (char*)malloc(sizeof(char) * (rows + 1));
				pic[l][rows] = '\0';
			}
			pic[columns] = (char*)malloc(sizeof(char) * 1);
			strcpy(pic[columns], "\0");
		}
		
		MPI_Type_vector(2, 1, 1, MPI_INT, &MPI_Answer);
		MPI_Type_commit(&MPI_Answer);
		

		MPI_Barrier(MPI_COMM_WORLD);
		Time = MPI_Wtime();
		
		/* Klassifikation */
		for (j = 0; j < 12; j++) {
			if (myRank == Master) {
				picptr = NextPic();
				PrintPic(picptr);
				
				/* Sende*/
				k = 0;
				for (l = 0; l < columns; l++) {
					for (m = 0; m < rows; m++) {
						buffer[k] = picptr[l][m];
						k++;
					} 
				}
				
				MPI_Bcast(buffer, size, MPI_CHAR, Master, MPI_COMM_WORLD);

				/* Empfange */
				MPI_Gather(Answer, 1, MPI_Answer, Answer, 1, MPI_Answer, Master, MPI_COMM_WORLD);
				
				MinCompare = MinCountChar = 1000;
				for (i = 1; i < nTasks; i++) {
					/* was reinschreiben */

					if (Answer[i].Compare < MinCompare) {
						MinCompare = Answer[i].Compare, MinVerIndex = (i - 1) % (TASK_NUM - 1);
					}
					
					if (Answer[i].CountChar < MinCountChar) {
						MinCountChar = Answer[i].CountChar, MinCountIndex = (i - 1) % (TASK_NUM - 1);
					}
				}
				
				#ifdef DEBUG
				if (MinCountIndex == MinVerIndex) {
					printf("Objekt %d ist ein %s\n", j, ObjectNames[MinVerIndex]);
				} else {
					printf("unterschiedliche Testergebnisse :\n\t nach Mustervergleich : %s\n\t nach Zeichen zaehlen : %s\n", ObjectNames[MinVerIndex], ObjectNames[MinCountIndex]);
				}
				#endif
			} else {
				/* Empfangen */
				MPI_Bcast(buffer, size, MPI_CHAR, Master, MPI_COMM_WORLD);
				k = 0;
				for (l = 0; l < columns; l++) {
					for (m = 0; m < rows; m++) {
						pic[l][m] = buffer[k];
						k++;
					} 
				}

				Answer[0].Compare = Compare(pic, pictures[me], 0, 7);
				Answer[0].CountChar = abs(CountChar(pic, 0, 7) - CountChar(pictures[me], 0, 7));

				/* Senden */
				MPI_Gather(Answer, 1, MPI_Answer, Answer, 1, MPI_Answer, Master, MPI_COMM_WORLD);
			}
		}
		
		MPI_Barrier(MPI_COMM_WORLD);
		DTime = MPI_Wtime();
		DTime -= Time;
		
		if (myRank == Master) {
			printf("Zeit %f\n", DTime);
		}
		
		
		free(buffer);
		MPI_Type_free(&MPI_Answer);
		
		if (myRank != Master) {
			for (l = 0; l < columns+1; l++) {
				free(pic[l]);
			}
			free(pic);
		}
		
	}
	
	MPI_Finalize();

	return 0;
}
