#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <mpi.h>
#include "functions.h"

#ifndef DEBUG
#define PrintPic(picptr) {}
#endif

#define Master 0
#define buffer_t char

#define TASK_NUM 1

MPI_Status status;
int myRank, nTasks;
buffer_t *buffer, *send_buffer;

int main(int Argc, char *Argv[])
{
	int *answer_compare;
	int *answer_countchar;
	int *answer_compare_local;
	int *answer_countchar_local;

	char **pic = dummy;
	char **picptr = NULL;
	char *ObjectNames[] = { "Kreis   ", "Dreieck ", "Rechteck"};

	int i = 0, j;
	int k, l, m;
	double DTime, Time;
	int columns, rows, size;
	int MinCompare, MinCountChar;
	int MinVerIndex, MinCountIndex;
	
	int *displ, *length;
	int lines, lines_extra;
	int first_line, last_line;

	MPI_Init(&Argc, &Argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &myRank);
	MPI_Comm_size(MPI_COMM_WORLD, &nTasks);
	
	rows = GetRows(pic);
	columns = GetColumns(pic);
	size = (rows * columns) + 1;

	srand(44444);
	if (nTasks < TASK_NUM || nTasks > columns) {
		if (myRank == Master) {
			printf("Programm sollte zwischen %d und %d Knoten laufen\n", TASK_NUM, columns);
		}
		MPI_Finalize();
		
		return 1;
	} else {
		lines = columns / nTasks;
		lines_extra = columns % nTasks;
		displ = (int*)malloc(sizeof(int) * nTasks);
		length = (int*)malloc(sizeof(int) * nTasks);
		
		for (i = 0; i < nTasks; i++) {
			if (i < lines_extra) {
				length[i] = lines + 1;
			} else {
				length[i] = lines;
			}
			
			length[i] *= rows;
		}
		
		j = 0;
		for (i = 0; i < nTasks; i++) {
			displ[i] = j;
			j += length[i];
		}
	
		buffer = (buffer_t*)malloc(length[myRank] * sizeof(buffer_t));
		send_buffer = (buffer_t*)malloc(size * sizeof(buffer_t));
		answer_compare = (int*)malloc(sizeof(int) * 3);
		answer_countchar = (int*)malloc(sizeof(int) * 3);
		answer_compare_local = (int*)malloc(sizeof(int) * 3);
		answer_countchar_local = (int*)malloc(sizeof(int) * 3);
		
		pic = (char**)malloc(sizeof(void*) * (columns + 1));
		for (l = 0; l < columns; l++) {
			pic[l] = (char*)malloc(sizeof(char) * (rows + 1));
			pic[l][rows] = '\0';
		}
		pic[columns] = (char*)malloc(sizeof(char) * 1);
		strcpy(pic[columns], "\0");
		

		MPI_Barrier(MPI_COMM_WORLD);
		Time = MPI_Wtime();
		
		/* Klassifikation */
		for (j = 0; j < 12; j++) {
			if (myRank == Master) {
				picptr = NextPic();
				PrintPic(picptr);
				
				/* Sende*/
				k = 0;
				for (l = 0; l < columns; l++) {
					for (m = 0; m < rows; m++) {
						send_buffer[k] = picptr[l][m];
						k++;
					} 
				}
			}
			/* Verteilen */
			MPI_Scatterv(send_buffer, length, displ, MPI_CHAR, buffer, length[myRank], MPI_CHAR, Master, MPI_COMM_WORLD);
			
			first_line = displ[myRank] / rows;
			last_line = (displ[myRank] + length[myRank]) / rows - 1;
			
			k = 0;
			for (l = first_line; l <= last_line; l++) {
				for (m = 0; m < rows; m++) {
					pic[l][m] = buffer[k];
					k++;
				} 
			}
			
			for (i = 0; i < 3; i++) {
				answer_compare_local[i] = Compare(pic, pictures[i], first_line, last_line);
				answer_countchar_local[i] = abs(CountChar(pic, first_line, last_line)
				                                - CountChar(pictures[i], first_line, last_line));
			}
			
			/* Zusammensuchen */
			MPI_Reduce(answer_compare_local, answer_compare, 3, MPI_FLOAT, MPI_SUM, Master, MPI_COMM_WORLD);
			MPI_Reduce(answer_countchar_local, answer_countchar, 3, MPI_FLOAT, MPI_SUM, Master, MPI_COMM_WORLD);


			if (myRank == Master) {
				MinCompare = MinCountChar = 1000;
				for (i = 0; i < 3; i++) {
					if (answer_compare[i] < MinCompare) {
						MinCompare = answer_compare[i], MinVerIndex = i;
					}
					
					if (answer_countchar[i] < MinCountChar) {
						MinCountChar = answer_countchar[i], MinCountIndex = i;
					}
				}
				
				#ifdef DEBUG
				if (MinCountIndex == MinVerIndex) {
					printf("Objekt %d ist ein %s\n", j, ObjectNames[MinVerIndex]);
				} else {
					printf("unterschiedliche Testergebnisse :\n\t nach Mustervergleich : %s\n\t nach Zeichen zaehlen : %s\n", ObjectNames[MinVerIndex], ObjectNames[MinCountIndex]);
				}
				#endif
			}
		}
		
		MPI_Barrier(MPI_COMM_WORLD);
		DTime = MPI_Wtime();
		DTime -= Time;
		
		if (myRank == Master) {
			printf("Zeit %f\n", DTime);
		}
		
		
		free(answer_compare);
		free(answer_countchar);
		free(answer_compare_local);
		free(answer_countchar_local);
		
		free(send_buffer);
		free(buffer);
		free(displ);
		free(length);
		
		for (l = 0; l < columns+1; l++) {
			free(pic[l]);
		}
		free(pic);
	}
	
	MPI_Finalize();

	return 0;
}
