#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <mpi.h>
#include "functions.h"

#define DEBUG 

#define AMOUNT 12

#ifndef DEBUG
#define PrintPic(picptr) {}
#endif

#define Master 0
#define buffer_t char

#define IDLE 0
#define WORKING 1
#define FINISHED 2

#define TASK_NUM 2

MPI_Status status;
int myRank, nTasks;
buffer_t* buffer;
buffer_t** global_buffer;
MPI_Request* requests;
int* finish_list;
MPI_Datatype MPI_Answer;
MPI_Status status;
int slaves;
int probe_flag;

int main(int Argc, char *Argv[])
{
	typedef struct {
		int Compare;
		int CountChar;
	}
	Answer_t;
	Answer_t *Answer;

	char **pic = dummy;
	char **picptr = NULL;
	char *ObjectNames[] = { "Kreis   ", "Dreieck ", "Rechteck"};

	int i = 0, j;
	int me;
	int k, l, m;
	double DTime, Time;
	int columns, rows, size;
	int MinCompare, MinCountChar;
	int MinVerIndex, MinCountIndex;
	int* answer;

	MPI_Init(&Argc, &Argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &myRank);
	MPI_Comm_size(MPI_COMM_WORLD, &nTasks);

	srand(44444);
	if (nTasks < TASK_NUM || nTasks > 13) {
		if (myRank == Master) {
			printf("Programm sollte zwischen %d und %d Knoten laufen\n", TASK_NUM, 13);
		}
		MPI_Finalize();
		
		return 1;
	} else {
		rows = GetRows(pic);
		columns = GetColumns(pic);
		size = (rows * columns) + 1;
		
		slaves = (nTasks - 1);
		
		if (myRank == Master) {
			answer = (int*)malloc(sizeof(int) * slaves);
			
			finish_list = (int*)malloc(sizeof(int) * slaves);
			for (i = 0; i < slaves; i++) {
				finish_list[i] = IDLE;
			}
			
			requests = (MPI_Request*)malloc(sizeof(MPI_Request) * slaves);
			
			global_buffer = (buffer_t**)malloc(sizeof(buffer_t*) * slaves);
			for (i = 0; i < slaves; i++) {
				global_buffer[i] = (buffer_t*)malloc(size * sizeof(buffer_t));
			}
		} else {
			Answer = (Answer_t*)malloc(sizeof(Answer_t) * 1);
			me = (myRank - 1) % (TASK_NUM - 1);
			
			buffer = (buffer_t*)malloc(size * sizeof(buffer_t));
			
			pic = (char**)malloc(sizeof(void*) * (columns + 1));
			for (l = 0; l < columns; l++) {
				pic[l] = (char*)malloc(sizeof(char) * (rows + 1));
				pic[l][rows] = '\0';
			}
			pic[columns] = (char*)malloc(sizeof(char) * 1);
			strcpy(pic[columns], "\0");
		}
		
		MPI_Type_vector(2, 1, 1, MPI_INT, &MPI_Answer);
		MPI_Type_commit(&MPI_Answer);
		

		MPI_Barrier(MPI_COMM_WORLD);
		Time = MPI_Wtime();
		
		/* Klassifikation */
		if (myRank == Master) {
			for (j = 0; j < AMOUNT; j ++) {
				picptr = NextPic();
				PrintPic(picptr);

				switch (finish_list[j % slaves]) {
				case WORKING:
					/* Wait until finished */
					if (finish_list[j % slaves] == WORKING) {
						MPI_Wait(&requests[j % slaves], &status); /* Request Daten freigeben */
						
						MPI_Recv(&answer[j % slaves], 1, MPI_INT, (j % slaves) + 1, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
						finish_list[j % slaves] = FINISHED;
					}
					
					/* jump to finished */
				case FINISHED:
					/* print answer */
					#ifdef DEBUG
					printf("Objekt %d ist ein %s\n", j, ObjectNames[answer[j % slaves]]);
					#endif
					
					finish_list[j % slaves] = IDLE;
					
					/* jump to idle */
				case IDLE:
					/* Sende*/
					k = 0;
					for (l = 0; l < columns; l++) {
						for (m = 0; m < rows; m++) {
							global_buffer[j % slaves][k] = picptr[l][m];
							k++;
						} 
					}
				
					MPI_Isend(global_buffer[j % slaves], size, MPI_CHAR, (j % slaves) + 1, j, MPI_COMM_WORLD, &requests[j % slaves]);
					finish_list[j % slaves] = WORKING;
				};
				
				do {
					MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &probe_flag, &status);
					
					if (probe_flag != 0) {
						MPI_Wait(&requests[status.MPI_SOURCE - 1], &status); /* Request Daten freigeben */
						
						MPI_Recv(&answer[status.MPI_SOURCE - 1], 1, MPI_INT, status.MPI_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
						finish_list[status.MPI_SOURCE - 1] = FINISHED;
					}
				} while (probe_flag != 0);
			}
			
			for (j = AMOUNT - slaves; j < AMOUNT; j ++) {
				switch (finish_list[j % slaves]) {
				case WORKING:
					/* Wait until finished */
					if (finish_list[j % slaves] == WORKING) {
						MPI_Wait(&requests[j % slaves], &status); /* Request Daten freigeben */
						
						MPI_Recv(&answer[j % slaves], 1, MPI_INT, (j % slaves) + 1, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
						finish_list[j % slaves] = FINISHED;
					}
					
					/* jump to finished */
				case FINISHED:
					/* print answer */
					#ifdef DEBUG
					printf("Objekt %d ist ein %s\n", j, ObjectNames[answer[j % slaves]]);
					#endif
					
					finish_list[j % slaves] = IDLE;
				case IDLE:
					break;
				};
			}
		} else {
			for (j = (myRank - 1) % slaves; j < AMOUNT; j += slaves) {
				/* Empfangen */
				MPI_Recv(buffer, size, MPI_CHAR, Master, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
				
				k = 0;
				for (l = 0; l < columns; l++) {
					for (m = 0; m < rows; m++) {
						pic[l][m] = buffer[k];
						k++;
					} 
				}
				
				MinCompare = MinCountChar = 1000;
				for (i = 0; i < 3; i++) {
					Answer[i].Compare = Compare(pictures[i], pic, 0, 7);
					Answer[i].CountChar = abs(CountChar(pic, 0, 7) - CountChar(pictures[i], 0, 7));
				
					if (Answer[i].Compare < MinCompare) {
						MinCompare = Answer[i].Compare, MinVerIndex = i;
					}
					
					if (Answer[i].CountChar < MinCountChar) {
						MinCountChar = Answer[i].CountChar, MinCountIndex = i;
					}
				}

				
				if (MinCountIndex == MinVerIndex) {
					answer = MinCountIndex;
				} else {
					#ifdef DEBUG
					printf("unterschiedliche Testergebnisse :\n\t nach Mustervergleich : %s\n\t nach Zeichen zaehlen : %s\n", ObjectNames[MinVerIndex], ObjectNames[MinCountIndex]);
					#endif
					answer = 0;
				}

				/* Senden */
				MPI_Send(&answer, 1, MPI_INT, Master, 0, MPI_COMM_WORLD);
			}
		}
		
		MPI_Barrier(MPI_COMM_WORLD);
		DTime = MPI_Wtime();
		DTime -= Time;
		
		if (myRank == Master) {
			printf("Zeit %f\n", DTime);
		}
		
		
		if (myRank == Master) {
			for (i = 0; i < slaves; i++) {
				free(global_buffer[i]);
			}
			free(global_buffer);
			
			free(finish_list);
			free(answer);
			free(requests);
		} else {
			free(buffer);
			free(Answer);
		}
		MPI_Type_free(&MPI_Answer);
		
		if (myRank != Master) {
			for (l = 0; l < columns+1; l++) {
				free(pic[l]);
			}
			free(pic);
		}
		
	}
	MPI_Finalize();

	return 0;
}
